from __future__ import annotations
from .PolynomialFraction import PolynomialFraction


class Expression:
    @staticmethod
    def fromString(string: str) -> Expression:
        pass

    def __init__(self, string): pass
