# from ...la1 import Matrix, MatrixOperation, MatrixOperationType


# def test_row_multiplication():
#     assert (Matrix([[5, 0], [0, 1]]), None) == MatrixOperation(MatrixOperationType.ROW_MULTIPLICATION,
#                                                                0, 5)(Matrix.id_matrix(2))


# def test_row_switching():
#     assert (Matrix([[0, 1], [1, 0]]), None) == MatrixOperation(
#         MatrixOperationType.ROW_SWITCHING, 0, 1)(Matrix.id_matrix(2))


# def test_row_addition():
#     assert (Matrix([[1, 1], [0, 1]]), None) == MatrixOperation(
#         MatrixOperationType.ROW_ADDITION, 0, 1)(Matrix.id_matrix(2))
